package kiev.stdlib;

import java.util.StringTokenizer;

public interface TypeInfoInterface {
	public TypeInfo getTypeInfoField();
}

public class TypeInfo {

	public static final TypeInfo[] emptyArray = new TypeInfo[0];

	private Class			type;
	private TypeInfo[]		typeargs;
	public  String			name;
	public  String			classname;
	
	/* Use java.util.Hashtable, since our own kiev.stdlib.Hashtable uses TypeInfo! */
	public static java.util.Hashtable typehash = new java.util.Hashtable(128);
	
	protected TypeInfo(TypeInfo[] ta, String nm, String clnm) {
		if( ta.length == 0 )
			typeargs = emptyArray;
		else
			typeargs = ta;
		name = nm.intern();
		classname = clnm.intern();
		typehash.put(nm,this);
//		System.out.println("TypeInfo for "+nm+" added");
	}
	
	protected TypeInfo(String nm, String clnm) {
		this(emptyArray,nm,clnm);
	}
	
	protected Class getClazz() {
		try {
			return type=Class.forName(classname);
		} catch( Throwable e ) {
			System.err.println("Error in Class.forName for "+classname);
			e.printStackTrace();
			return null;
//			throw new RuntimeException(e.getMessage());
		}
	}
	
	public static TypeInfo newTypeInfo(String t) {
		try {
			t = t.intern();
			if( typehash == null )
				typehash = new java.util.Hashtable(128);
			TypeInfo ti = (TypeInfo)typehash.get(t);
			if( ti != null ) return ti;
			if( t.charAt(0) == '[' ) {
				return new TypeInfo(t,t);
			}
			int indx = t.indexOf('<');
			if( indx < 0 ) {
				return new TypeInfo(t,t);
			}
			String targs = t.substring(indx+1,t.length()-1);
			String t1 = t.substring(0,indx).intern();
			TypeInfo[] args = new TypeInfo[0];
			// Split targs into args array
			char c;
			int len = targs.length();
			int start_pos = 0;
			for(int i=0; i < len; i++) {
				c = targs.charAt(i);
				if( c == ',' ) {
					args = (TypeInfo[])Arrays.append(args,newTypeInfo(targs.substring(start_pos,i)));
					start_pos = i+1;
				}
				else if( c == '<' ) {
					// Scan until matched '>'
					int depth = 1;
					for(i++; depth > 0; i++) {
						c = targs.charAt(i);
						if( c == '>' ) depth--;
						else if( c == '<' ) depth++;
					}
					args = (TypeInfo[])Arrays.append(args,newTypeInfo(targs.substring(start_pos,i)));
					start_pos = i;
				}
			}
			if( start_pos < len-1 )
				args = (TypeInfo[])Arrays.append(args,newTypeInfo(targs.substring(start_pos,len)));
			return new TypeInfo(args,t,t1);
		} catch( Throwable e ) {
			e.printStackTrace();
			return null;
		}
	}

	public String toString() {
		return name;
//		if( typeargs == emptyArray ) return type.toString();
//		StringBuffer sb = new StringBuffer(128);
//		sb.append(type).append('<');
//		for(int i=0; i < typeargs.length; i++) {
//			sb.append(typeargs[i]);
//			if( i < typeargs.length-1 )
//				sb.append(',');
//		}
//		sb.append('>');
//		return sb.toString();
	}
	
	public boolean $instanceof(Object obj) {
//		System.out.println("Checking '"+obj+"' of class "+obj.getClass()+" for "+this);
		if( obj == null ) return true;
		try {
			if( !type.isInstance(obj) ) return false;
		} catch( NullPointerException e ) {
			if( !getClazz().isInstance(obj) ) return false;
		}
		if( obj instanceof TypeInfoInterface ) {
			TypeInfo ti = ((TypeInfoInterface)obj).getTypeInfoField();
			return this.$instanceof(ti);
		} else {
			return typeargs == emptyArray;
		}
	}

	public boolean $instanceof(Object obj, TypeInfo oti) {
//		System.out.println("Checking '"+obj+"' of type "+oti+" for "+this);
		if( obj == null ) return true;
		try {
			if( !type.isInstance(obj) ) return false;
		} catch( NullPointerException e ) {
			if( !getClazz().isInstance(obj) ) return false;
		}
		if( typeargs.length != oti.typeargs.length ) return false;
//		System.out.println("Class and number of args ("+typeargs.length+") match, cheking each arg");
		for(int i=0; i < typeargs.length; i++) {
			if( !typeargs[i].$ti_instanceof_ti(oti.typeargs[i]) ) return false;
		}
//		System.out.println("Type of object '"+obj+"' of type "+oti+" match type "+this);
		return true;
	}

	public boolean $ti_instanceof_ti(TypeInfo oti) {
//		System.out.println("Checking type "+oti+" for "+this);
		if( name != oti.name ) {
			if( type == null ) getClazz();
			if( oti.type == null ) oti.getClazz();
			if( !type.isAssignableFrom(oti.type) ) return false;
			if( typeargs.length != oti.typeargs.length ) return false;
			for(int i=0; i < typeargs.length; i++) {
				if( !typeargs[i].$ti_instanceof_ti(oti.typeargs[i]) ) return false;
			}
		}
//		System.out.println("Type "+oti+" match type "+this);
		return true;
	}
	
	public Object newArray(int i) {
		try {
			return java.lang.reflect.Array.newInstance(type,i);
		} catch( NullPointerException e ) {
			return java.lang.reflect.Array.newInstance(getClazz(),i);
		}
	}

	public Object newArray(int[] i) {
		try {
			return java.lang.reflect.Array.newInstance(type,i);
		} catch( NullPointerException e ) {
			return java.lang.reflect.Array.newInstance(getClazz(),i);
		}
	}

	public Object newArray(int arg, int i) {
		return typeargs[arg].newArray(i);
	}

	public Object newArray(int arg, int[] i) {
		return typeargs[arg].newArray(i);
	}

}


public class TypeInfoTest {

	public static void main(String[] args) {
		TypeInfo s = TypeInfo.newTypeInfo("java.lang.String");
		TypeInfo i = TypeInfo.newTypeInfo("java.lang.Integer");
		TypeInfo ls = TypeInfo.newTypeInfo("kiev.stdlib.List<java.lang.String>");
		TypeInfo li = TypeInfo.newTypeInfo("kiev.stdlib.List<java.lang.Integer>");
		TypeInfo jht = TypeInfo.newTypeInfo("java.util.Hashtable");
		TypeInfo llks = TypeInfo.newTypeInfo("kiev.stdlib.List<kiev.stdlib.List<kiev.stdlib.KString>>");
		TypeInfo ht_s_llks = TypeInfo.newTypeInfo("kiev.stdlib.Hashtable<java.lang.String,kiev.stdlib.List<kiev.stdlib.List<kiev.stdlib.KString>>>");
		System.out.println("----------");
		System.out.println("typehash = "+TypeInfo.typehash);
		System.out.println("----------");
		System.out.println("String instanceof "+s+" == "+s.$instanceof("hello"));
		System.out.println("----------");
		System.out.println("Hashtable instanceof "+jht+" == "+jht.$instanceof(new java.util.Hashtable()));
		System.out.println("----------");
		System.out.println("List<List<KString>> instanceof "+llks+" == "+llks.$instanceof(new List.Cons<List<KString>>(new List.Cons<KString>(KString.Empty,List.Nil),List.Nil),llks));	
		System.out.println("----------");
		System.out.println("String instanceof "+i+" == "+i.$instanceof("hello"));
		System.out.println("----------");
		System.out.println("List<String> instanceof "+ls+" == "+ls.$instanceof(new List.Cons<String>("hello",List.Nil),ls));
		System.out.println("----------");
		System.out.println("List<Integer> instanceof "+li+" == "+li.$instanceof(new List.Cons<Integer>(new Integer(1),List.Nil),li));
		System.out.println("----------");
		System.out.println("List<String> instanceof "+li+" == "+li.$instanceof(new List.Cons<String>("hello",List.Nil),ls));
		System.out.println("----------");
		System.out.println("List<Integer> instanceof "+ls+" == "+ls.$instanceof(new List.Cons<Integer>(new Integer(1),List.Nil),li));
	}

}
